@@ -12,38 +12,208 @@
 #
 # \arguments{
 #   \item{B}{A postive @integer specifying the number of bootstrap samples.}
+#   \item{boot}{Alternatively, to generating \code{B} bootstrap samples,
+#      this specifies a pre-generated set of bootstrap samples as
+#      returned by \code{bootstrapSegmentsAndChangepoints()}.}
+#   \item{...}{Additional arguments passed to \code{bootstrapSegmentsAndChangepoints()}.}
 #   \item{probs}{The default quantiles to be estimated.}
 #   \item{statsFcn}{A (optional) @function that estimates confidence
 #      intervals given locus-level data.
 #      If @NULL, the @see "stats::quantile" function is used.}
-#   \item{by}{A @character specifying whether DH should be calculated from
-#      normalized ('betaTN') or non-normalized ('betaT') tumor BAFs.}
+#   \item{what}{A @character @vector specifying what to bootstrap.}
 #   \item{force}{If @TRUE, already existing estimates are ignored,
 #      otherwise not.}
-#   \item{seed}{(optional) A random seed.}
 #   \item{verbose}{See @see "R.utils::Verbose".}
 #   \item{.debug}{(internal) If @TRUE, additional sanity checks are
-#      performed internally..}
-#   \item{...}{Not used.}
+#      performed internally.}
 # }
 #
 # \value{
-#   Returns a @see "PairedPSCBS" object with copy-neutral calls.
+#   Returns a @see "PairedPSCBS" object.
 # }
 #
 # @author "HB"
-#
-# \seealso{
-#   Internally, one of the following methods are used:
-#   @seemethod "callCopyNeutralByTCNofAB".
-# }
-#
 #*/###########################################################################
-setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, probs=c(0.025, 0.050, 0.95, 0.975), statsFcn=NULL, by=c("betaTN", "betaT"), force=FALSE, seed=NULL, verbose=FALSE, .debug=FALSE, ...) {
+setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, boot=NULL, ..., probs=c(0.025, 0.050, 0.95, 0.975), statsFcn=NULL, what=c("segment", "changepoint"), force=FALSE, verbose=FALSE, .debug=FALSE) {
   # Settings for sanity checks
   tol <- getOption("PSCBS/sanityChecks/tolerance", 0.0005);
 
 
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Local functions
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  summarizeSamples <- function(X, statsFcn, stats=NULL, what=c("segment", "changepoint"), ..., verbose=FALSE) {
+    # Argument 'X':
+    stopifnot(is.array(X));
+    dim <- dim(X);
+    stopifnot(length(dim) == 3L);
+
+    # Argument 'statsFcn':
+    stopifnot(is.function(statsFcn));
+    statsT <- statsFcn(1);
+    stopifnot(!is.null(names(statsT)));
+    nbrOfStats <- length(statsT);
+    statsNames <- names(statsT);
+    statsT <- NULL; # Not needed anymore
+
+    # Argument 'stats':
+    if (!is.null(stats)) {
+      stopifnot(is.data.frame(stats));
+    }
+
+    # Argument 'what':
+    what <- match.arg(what);
+    whatC <- capitalize(what);
+
+    # Argument 'verbose':
+    verbose <- Arguments$getVerbose(verbose);
+    if (verbose) {
+      pushState(verbose);
+      on.exit(popState(verbose));
+    }
+
+
+    dimnames <- dimnames(X);
+    fields <- dimnames[[3L]];
+
+    verbose && enter(verbose, sprintf("Summarizing bootstrapped %s (%s) data", what, paste(sQuote(fields), collapse=", ")));
+
+    # Allocate JxQxF matrix S
+    dim[2L] <- nbrOfStats;
+    dimnames[[2L]] <- statsNames;
+    S <- array(NA_real_, dim=dim, dimnames=dimnames);
+    verbose && str(verbose, S);
+
+    for (kk in seq(along=fields)) {
+      field <- fields[kk];
+      verbose && enter(verbose, sprintf("Field #%d ('%s') of %d", kk, field, length(fields)));
+
+      Xkk <- X[,,kk,drop=FALSE];  # An JxB matrix
+      dim(Xkk) <- dim(Xkk)[-3L];
+      # Sanity check
+      stopifnot(is.matrix(Xkk));
+      stopifnot(nrow(Xkk) == dim(X)[1L]);
+      stopifnot(ncol(Xkk) == B);
+
+      for (jj in seq(length=dim(X)[1L])) {
+        verbose && enter(verbose, sprintf("%s #%d of %d", whatC, jj, dim(X)[1L]));
+        Xkkjj <- Xkk[jj,,drop=TRUE]; # A vector of length B
+        S[jj,,kk] <- statsFcn(Xkkjj);
+        verbose && exit(verbose);
+      } # for (jj ...)
+
+      Xkk <- NULL; # Not needed anymore
+
+      verbose && exit(verbose);
+    } # for (jj ...)
+
+    # Not needed anymore
+    X <- NULL;
+
+    verbose && cat(verbose, "Bootstrap statistics");
+    verbose && str(verbose, S);
+
+    # Reshape JxQx4 array to Jx(4*Q) matrix
+    T <- wrap(S, map=list(1,NA), sep="_");
+    colnames(T) <- gsub("(.*)_(.*)", "\\2_\\1", colnames(T));
+
+    # Append as new columns to the summary table
+    stats <- cbind(stats, T);
+
+    # Drop previously estimated values
+    dups <- duplicated(colnames(stats), fromLast=TRUE);
+    if (any(dups)) {
+      stats <- stats[,!dups, drop=FALSE];
+    }
+
+    # Not needed anymore
+    T <- dups <- NULL;
+
+
+    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+    # Statistical sanity checks
+    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+    if (what == "segment" && B >= 100L) {
+      verbose && enter(verbose, "Statistical sanity checks (iff B >= 100)");
+
+      stopifnot(is.array(S));
+
+      # Find extreme quantiles
+      probs <- dimnames(S)[[2L]];
+      verbose && printf(verbose, "Available summaries: %s\n", paste(probs, collapse=", "));
+      probs <- grep("%", probs, fixed=TRUE, value=TRUE);
+      S <- S[,probs,,drop=FALSE];
+      probs <- gsub("%", "", probs, fixed=TRUE);
+      probs <- as.double(probs) / 100;
+      verbose && printf(verbose, "Available quantiles: %s\n", paste(probs, collapse=", "));
+      verbose && str(verbose, S);
+      # Sanity check
+      stopifnot(all(is.finite(probs)));
+
+      # Is it possible to check?
+      if (any(probs < 0.10) && any(probs > 0.90)) {
+        tryCatch({
+          fields <- dimnames(S)[[3L]];
+          for (kk in seq(along=fields)) {
+            field <- fields[kk];
+            verbose && enter(verbose, sprintf("Field #%d ('%s') of %d", kk, field, length(fields)));
+
+            # Bootstrap statistics
+            Skk <- S[,,kk, drop=FALSE];
+            dim(Skk) <- dim(Skk)[-3L];
+
+            # Sanity checks
+            stopifnot(is.matrix(Skk));
+
+            range <- Skk[,c(1L,ncol(Skk)),drop=FALSE];
+
+            # Segmentation means
+            key <- sprintf("%sMean", field);
+            segMean <- segs[[key]];
+
+            # Segmentation counts
+            cfield <- sprintf("%sNbrOfLoci", ifelse(field == "tcn", "tcn", "dh"));
+            counts <- segs[,cfield,drop=TRUE];
+
+            if (verbose) {
+              for (rr in seq_len(length(segMean))) {
+                printf(verbose, "Seg %3d. mean=%g, range=[%g,%g], n=%d\n", rr, segMean[rr], range[rr,1L], range[rr,2L], counts[rr]);
+              } # for (rr ...)
+            }
+
+            # Compare only segments with enough data points
+            keep <- (counts > 1L);
+            range <- range[keep,,drop=FALSE];
+            segMean <- segMean[keep];
+
+            # Sanity checks
+            stopifnot(all(range[,2L] + tol >= range[,1L], na.rm=TRUE));
+            stopifnot(all(segMean + tol >= range[,1L], na.rm=TRUE));
+            stopifnot(all(segMean - tol <= range[,2L], na.rm=TRUE));
+
+            verbose && exit(verbose);
+          } # for (kk ...)
+        }, error = function(ex) {
+          # If an error, display the data, then throw the exception
+          verbose && cat(verbose, "Tolerance (option 'PSCBS/sanityChecks/tolerance'): ", tol);
+          verbose && print(verbose, segs);
+          throw(ex);
+        })
+      } else {
+        verbose && cat(verbose, "Skipping. Not enough quantiles: ",
+                                 paste(dimnames(S)[[2L]], collapse=", "));
+      }
+
+      verbose && exit(verbose);
+    } # if (B >= 100L)
+
+
+    verbose && exit(verbose);
+
+    stats;
+  } # summarizeSamples()
+
+
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   # Validate arguments
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -61,16 +231,8 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
     statsFcn <- function(x) quantile(x, probs=probs, na.rm=TRUE);
   }
 
-  # Argument 'by':
-  by <- match.arg(by);
-
-  # Argument 'seed':
-  if (!is.null(seed)) {
-    seed <- Arguments$getInteger(seed);
-  }
-
-  # Argument '.debug':
-  .debug <- Arguments$getLogical(.debug);
+  # Argument 'what':
+  what <- unique(match.arg(what, several.ok=TRUE));
 
   # Argument 'force':
   force <- Arguments$getLogical(force);
@@ -82,9 +244,162 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
     on.exit(popState(verbose));
   }
 
+  # Argument '.debug':
+  .debug <- Arguments$getLogical(.debug);
 
 
-  verbose && enter(verbose, "Resample (TCN,DH) signals and re-estimate mean levels");
+
+  verbose && enter(verbose, "Resample (TCN,DH) signals and re-estimate summaries for ", paste(what, collapse=" & "));
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Extract existing estimates
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  if (is.element("segment", what)) {
+    segs <- getSegments(fit);
+  }
+  if (is.element("changepoint", what)) {
+    cps <- getChangePoints(fit);
+  }
+
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Already done?
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  stats <- statsFcn(1);
+  stopifnot(!is.null(names(stats)));
+  nbrOfStats <- length(stats);
+  statsNames <- names(stats);
+
+  if (is.element("segment", what)) {
+    tcnStatsNames <- sprintf("tcn_%s", names(stats));
+    dhStatsNames <- sprintf("dh_%s", names(stats));
+    c1StatsNames <- sprintf("c1_%s", names(stats));
+    c2StatsNames <- sprintf("c2_%s", names(stats));
+    allStatsNames <- c(tcnStatsNames, dhStatsNames, c1StatsNames, c2StatsNames);
+    isDone <- is.element(allStatsNames, names(segs));
+    names(isDone) <- allStatsNames;
+    verbose && cat(verbose, "Already done?");
+    verbose && print(verbose, isDone);
+
+    # Not needed anymore
+    allStatsNames <- tcnStatsNames <- dhStatsNames <-
+                     c1StatsNames <- c2StatsNames <- NULL;
+
+    if (!force && all(isDone)) {
+      verbose && cat(verbose, "Already done. Skipping.");
+      verbose && exit(verbose);
+      return(fit);
+    }
+  }
+
+
+  # The object to be returned
+  fitB <- fit;
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Bootstrap (TCN,DH,C1,C2) segment mean levels
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  if (is.null(boot)) {
+    boot <- bootstrapSegmentsAndChangepoints(fit, B=B, ...,
+                         force=force, .debug=.debug, verbose=verbose);
+  } else {
+    B <- dim(boot$segments)[2L];
+  }
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Summarizing segment (TCN,DH,C1,C2) mean levels
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  if (is.element("segment", what)) {
+    segs <- summarizeSamples(boot$segments, statsFcn=statsFcn, stats=segs, what="segment", verbose=verbose);
+    # Record statistics
+    fitB$output <- segs;
+    segs <- NULL; # Not needed anymore
+  }
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Summarizing change point (alpha, radius, manhattan, d1, d2) data
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  if (is.element("changepoint", what)) {
+    cps <- summarizeSamples(boot$changepoints, statsFcn=statsFcn, stats=cps, what="changepoint", verbose=verbose);
+    # Record statistics
+    fitB$changepoints <- cps;
+    cps <- NULL; # Not needed anymore
+  }
+
+  # Not needed anymore
+  fit <- boot <- NULL;
+
+  verbose && exit(verbose);
+
+  fitB;
+}, private=TRUE) # bootstrapTCNandDHByRegion()
+
+
+
+
+
+#   \item{by}{A @character specifying whether DH should be calculated from
+#      normalized ('betaTN') or non-normalized ('betaT') tumor BAFs.}
+#   \item{seed}{(optional) A random seed.}
+#
+#
+# \value{
+#   Returns a named @list containing two @arrays of bootstrap samples.
+#   These arrays also contains the original observation as the first
+#   element before the actual bootstrap samples.
+# }
+setMethodS3("bootstrapSegmentsAndChangepoints", "PairedPSCBS", function(fit, B=1000L, by=c("betaTN", "betaT"), seed=NULL, force=FALSE, cache=FALSE, verbose=FALSE, .debug=FALSE, ...) {
+  # Settings for sanity checks
+  tol <- getOption("PSCBS/sanityChecks/tolerance", 0.0005);
+
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Validate arguments
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Argument 'B':
+  B <- Arguments$getInteger(B, range=c(1,Inf));
+
+  # Argument 'by':
+  by <- match.arg(by);
+
+  # Argument 'seed':
+  if (!is.null(seed)) {
+    seed <- Arguments$getInteger(seed);
+  }
+
+  # Argument '.cache':
+  cache <- Arguments$getLogical(cache);
+
+  # Argument 'force':
+  force <- Arguments$getLogical(force);
+
+  # Argument 'verbose':
+  verbose <- Arguments$getVerbose(verbose);
+  if (verbose) {
+    pushState(verbose);
+    on.exit(popState(verbose));
+  }
+
+  # Argument '.debug':
+  .debug <- Arguments$getLogical(.debug);
+
+
+
+  verbose && enter(verbose, "Bootstrapping (TCN,DH,C1,C2) segment mean levels");
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Check for cached results
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  key <- list(method="bootstrapSegmentsAndChangepoints", class=class(fit)[1L],
+              fit=fit, B=B, by=by, seed=seed);
+  dirs <- c("PSCBS", "bootstrap");
+  boot <- loadCache(key=key, dirs=dirs);
+  if (!force && !is.null(boot)) {
+    verbose && cat(verbose, "Found cached results. Skipping.");
+    verbose && exit(verbose);
+    return(boot);
+  }
+
 
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   # Set the random seed
@@ -123,7 +438,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
   if (!params$joinSegments) {
     throw("Cannot bootstrap TCN and DH by segments unless PSCNs are segmented using joinSegments=TRUE.");
   }
-  if (regexpr(",", params$flavor, fixed=TRUE) != -1) {
+  if (regexpr(",", params$flavor, fixed=TRUE) != -1L) {
     throw(sprintf("Cannot bootstrap TCN and DH by segments if PSCNs are segmented using flavor=\"%s\".", params$flavor));
   }
   # Sanity check (same as above, but just in case)
@@ -133,35 +448,13 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
 
 
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  # Find estimates to be done
+  # Find estimators
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  stats <- statsFcn(1);
-  stopifnot(!is.null(names(stats)));
-  nbrOfStats <- length(stats);
-  statsNames <- names(stats);
-
-  # Already done?
-  tcnStatsNames <- sprintf("tcn_%s", names(stats));
-  dhStatsNames <- sprintf("dh_%s", names(stats));
-  c1StatsNames <- sprintf("c1_%s", names(stats));
-  c2StatsNames <- sprintf("c2_%s", names(stats));
-  allStatsNames <- c(tcnStatsNames, dhStatsNames, c1StatsNames, c2StatsNames);
-  isDone <- is.element(allStatsNames, names(segs));
-  names(isDone) <- allStatsNames;
-  verbose && cat(verbose, "Already done?");
-  verbose && print(verbose, isDone);
-
-  if (!force && all(isDone)) {
-    verbose && cat(verbose, "Already done. Skipping.");
-    verbose && exit(verbose);
-    return(fit);
-  }
-
-
-  # Get mean estimators
+  # Get mean estimators used
   estList <- getMeanEstimators(fit, c("tcn", "dh"));
   avgTCN <- estList$tcn;
   avgDH <- estList$dh;
+  estList <- NULL; # Not needed anymore
 
 
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -194,7 +487,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
   verbose && cat(verbose, "Number of non-SNPs: ", nbrOfNonSNPs);
 
   # Sanity checks
-  stopifnot(length(intersect(snps, nonSNPs)) == 0);
+  stopifnot(length(intersect(snps, nonSNPs)) == 0L);
 
   # Heterozygous SNPs
   isHet <- isSNP & (muN == 1/2);
@@ -208,7 +501,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
                                       nbrOfHoms, 100*nbrOfHoms/nbrOfSNPs);
 
   # Sanity checks
-  stopifnot(length(intersect(hets, homs)) == 0);
+  stopifnot(length(intersect(hets, homs)) == 0L);
   stopifnot(nbrOfHets + nbrOfHoms == nbrOfSNPs);
 
   # Sanity checks
@@ -242,10 +535,9 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
   nbrOfSegments <- nrow(segs);
 
   # Allocate JxBx4 matrix M of bootstrap means
-  naValue <- as.double(NA);
-  dim <- c(nbrOfSegments, B, 4);
+  dim <- c(nbrOfSegments, B, 4L);
   dimnames <- list(NULL, NULL, c("tcn", "dh", "c1", "c2"));
-  M <- array(naValue, dim=dim, dimnames=dimnames);
+  M <- array(NA_real_, dim=dim, dimnames=dimnames);
   verbose && str(verbose, M);
 
   # Identify all loci with non-missing signals
@@ -253,14 +545,14 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
   idxsRho <- which(!is.na(rho));
 
   # Vectorized pre-adjustments
-  for (key in c("tcnNbrOfLoci", "dhNbrOfLoci")) {
-    counts <- segs[[key]];
+  for (field in c("tcnNbrOfLoci", "dhNbrOfLoci")) {
+    counts <- segs[[field]];
     counts[is.na(counts)] <- 0L;
-    segs[[key]] <- counts;
+    segs[[field]] <- counts;
   }
 
-  hasTcnLoci <- (is.finite(tcnSegRows[,1]) & is.finite(tcnSegRows[,2]));
-  hasDhLoci <- (is.finite(dhSegRows[,1]) & is.finite(dhSegRows[,2]));
+  hasTcnLoci <- (is.finite(tcnSegRows[,1L]) & is.finite(tcnSegRows[,2L]));
+  hasDhLoci <- (is.finite(dhSegRows[,1L]) & is.finite(dhSegRows[,2L]));
 
   # Identify "splitter" segments which have no data
   chrs <- segs[["chromosome"]];
@@ -268,7 +560,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
   dhIds <- segs[["dhId"]];
   tcnMeans <- segs[["tcnMean"]];
   dhMeans <- segs[["dhMean"]];
-  isSplitter <- (is.na(chrs) && is.na(tcnIds) && is.na(dhIds));
+  isSplitter <- (is.na(chrs) & is.na(tcnIds) & is.na(dhIds));
 
   # Get all segment indices except for "splitters"
   jjs <- seq(length=nbrOfSegments);
@@ -302,9 +594,9 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
 
     # Indices of all loci
     if (hasTcnLoci[jj]) {
-      idxsAll <- tcnSegRowJJ[1]:tcnSegRowJJ[2];
+      idxsAll <- tcnSegRowJJ[1L]:tcnSegRowJJ[2L];
     } else {
-      idxsAll <- integer(0);
+      idxsAll <- 0L;
     }
 
     verbose && str(verbose, idxsAll);
@@ -330,12 +622,12 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
     verbose && enter(verbose, "Identify loci used to bootstrap DH means");
 
     if (hasDhLoci[jj]) {
-      idxsDH <- dhSegRowJJ[1]:dhSegRowJJ[2];
+      idxsDH <- dhSegRowJJ[1L]:dhSegRowJJ[2L];
       idxsDH <- intersect(idxsDH, hets);
       # Drop missing values
       idxsDH <- intersect(idxsDH, idxsRho);
     } else {
-      idxsDH <- integer(0);
+      idxsDH <- 0L;
     }
 
     verbose && cat(verbose, "Heterozygous SNPs to resample for DH:");
@@ -392,7 +684,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
     # Sanity check
     if (.debug) {
       stopifnot(length(idxsHet) + length(idxsHom) + length(idxsNonSNP) == nbrOfTCNs);
-      stopifnot(length(intersect(idxsDH, idxsHetNonDH)) == 0);
+      stopifnot(length(intersect(idxsDH, idxsHetNonDH)) == 0L);
       stopifnot(length(idxsTCN) == nbrOfTCNs);
     }
 
@@ -414,7 +706,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
       dMu <- (mu - tcnMeans[jj]);
       if (abs(dMu) > tol) {
         str(list(nbrOfTCNs=nbrOfTCNs, tcnNbrOfLoci=segJJ$tcnNbrOfLoci, mu=mu, tcnMean=tcnMeans[jj], dMu=dMu, "abs(dMu)"=abs(dMu), "range(x[units])"=range(x[idxsTCN])));
-        throw(sprintf("INTERNAL ERROR: Incorrect recalculated TCN mean for Segment #%d (chr %d, tcnId=%d, dhId=%d): %g != %g", jj, chr, tcnId, dhId, mu, tcnMeans[jj]));
+        throw(sprintf("INTERNAL ERROR: Incorrectly recalculated TCN mean for Segment #%d (chr %d, tcnId=%d, dhId=%d): %g != %g", jj, chr, tcnId, dhId, mu, tcnMeans[jj]));
       }
     }
 
@@ -426,7 +718,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
       dMu <- (mu - dhMeans[jj]);
       if (abs(dMu) > tol) {
         str(list(nbrOfDHs=nbrOfDHs, dhNbrOfLoci=segJJ$dhNbrOfLoci, mu=mu, dhMean=dhMeans[jj], dMu=dMu, "abs(dMu)"=abs(dMu), "range(x[units])"=range(x[idxsDH])));
-        throw(sprintf("INTERNAL ERROR: Incorrect recalculated DN mean for Segment #%d (chr %d, tcnId=%d, dhId=%d): %g != %g", jj, chr, tcnId, dhId, mu, dhMeans[jj]));
+        throw(sprintf("INTERNAL ERROR: Incorrectly recalculated DH mean for Segment #%d (chr %d, tcnId=%d, dhId=%d): %g != %g", jj, chr, tcnId, dhId, mu, dhMeans[jj]));
       }
     }
 
@@ -497,7 +789,7 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
     verbose && exit(verbose);
   } # for (jj ...)
 
-  verbose && cat(verbose, "Bootstrap means");
+  verbose && cat(verbose, "Bootstrapped segment mean levels");
   verbose && str(verbose, M);
 
   # Sanity check
@@ -505,161 +797,162 @@ setMethodS3("bootstrapTCNandDHByRegion", "PairedPSCBS", function(fit, B=1000L, p
 
 
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  # Calculate (C1,C2) bootstrap statistics
+  # Add (C1,C2) bootstrap mean levels
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  verbose && enter(verbose, "Calculating (C1,C2) from (TCN,DH) bootstraps");
-  C1 <- (1-M[,,"dh"]) * M[,,"tcn"] / 2;
-  C2 <- M[,,"tcn"] - C1;
+  verbose && enter(verbose, "Calculating (C1,C2) mean levels from (TCN,DH) mean levels");
+  C1 <- (1-M[,,"dh", drop=FALSE]) * M[,,"tcn", drop=FALSE] / 2;
+  C2 <- M[,,"tcn", drop=FALSE] - C1;
   M[,,"c1"] <- C1;
   M[,,"c2"] <- C2;
   verbose && str(verbose, M);
+  # Sanity check
+  stopifnot(dim(M)[1L] == nbrOfSegments);
+  stopifnot(all(!is.nan(M)));
+  # Not needed anymore
+  C1 <- C2 <- NULL;
+  verbose && exit(verbose);
+
+
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Bootstrap polar (alpha,radius,manhattan) for change points
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  verbose && enter(verbose, "Calculating polar (alpha,radius,manhattan) for change points");
+  C <- M[,,c("c1","c2"), drop=FALSE];
+
+  # Calculate difference
+  # (will be empty if nbrOfSegments == 1, but that's ok/intended)
+  D <- C[-nbrOfSegments,,, drop=FALSE] - C[-1L,,, drop=FALSE];
+  # Sanity check
+  stopifnot(dim(D)[1L] == nbrOfSegments-1L);
+  stopifnot(all(!is.nan(D)));
+  C <- NULL; # Not needed anymore
+
+  # Allocate array
+  dimnames <- dimnames(D);
+  dimnames[[3L]] <- c("alpha", "radius", "manhattan", "d1", "d2");
+  dim <- dim(D);
+  dim[3L] <- length(dimnames[[3L]]);
+  P <- array(NA_real_, dim=dim, dimnames=dimnames);
+  stopifnot(dim(P)[1L] == nbrOfSegments-1L);
+
+  if (nbrOfSegments >= 2L) {
+    verbose && str(verbose, D);
+    P[,,"alpha"] <- atan2(D[,,2], D[,,1]); # Changepoint angles in (0,2*pi)
+    P[,,"radius"] <- sqrt(D[,,2]^2 + D[,,1]^2);
+    P[,,"manhattan"] <- abs(D[,,2]) + abs(D[,,1]);
+    P[,,"d1"] <- D[,,1];
+    P[,,"d2"] <- D[,,2];
+  }
+  alpha <- D <- NULL; # Not needed anymore
+  verbose && cat(verbose, "Bootstrapped change points");
+  verbose && str(verbose, P);
 
   # Sanity check
-  stopifnot(all(!is.nan(M)));
+  stopifnot(dim(P)[1L] == nbrOfSegments-1L);
+  stopifnot(all(!is.nan(P)));
   verbose && exit(verbose);
 
+  boot <- list(segments=M, changepoints=P);
 
-  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  # Calculate bootstrap statistics
-  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  verbose && enter(verbose, "Calculating (TCN,DH) bootstrap statistics");
-  # Allocate JxQx4 matrix S
-  naValue <- as.double(NA);
-  dim <- dim(M);
-  dimnames <- dimnames(M);
-  dim[2] <- nbrOfStats;
-  dimnames[[2]] <- statsNames;
-  S <- array(naValue, dim=dim, dimnames=dimnames);
-  verbose && str(verbose, S);
-
-  fields <- dimnames(M)[[3]];
-  for (kk in seq(along=fields)) {
-    field <- fields[kk];
-    verbose && enter(verbose, sprintf("Field #%d ('%s') of %d", kk, field, length(fields)));
-
-    Mkk <- M[,,kk,drop=FALSE];  # An JxB matrix
-    dim(Mkk) <- dim(Mkk)[-3];
-    # Sanity check
-    stopifnot(is.matrix(Mkk));
-    stopifnot(nrow(Mkk) == nbrOfSegments);
-    stopifnot(ncol(Mkk) == B);
-
-    for (jj in seq(length=nbrOfSegments)) {
-      verbose && enter(verbose, sprintf("Segment #%d of %d", jj, nbrOfSegments));
-
-      Mkkjj <- Mkk[jj,,drop=TRUE]; # A vector of length B
-
-      S[jj,,kk] <- statsFcn(Mkkjj);
-
-      verbose && exit(verbose);
-    } # for (jj ...)
-
-    verbose && exit(verbose);
-  } # for (jj ...)
-  verbose && exit(verbose);
-  verbose && cat(verbose, "Bootstrap statistics");
-  verbose && str(verbose, S);
-
-
-
-  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  # Store
-  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  # Reshape JxQx4 array to Jx(4*Q) matrix
-  T <- wrap(S, map=list(1,NA), sep="_");
-  colnames(T) <- gsub("(.*)_(.*)", "\\2_\\1", colnames(T));
-
-  # Append
-  segs <- cbind(segs, T);
-
-  # Drop previously estimated values
-  dups <- duplicated(colnames(segs), fromLast=TRUE);
-  if (any(dups)) {
-    segs <- segs[,!dups, drop=FALSE];
+  # Cache?
+  if (cache) {
+    saveCache(boot, key=key, dirs=dirs);
+    verbose && cat(verbose, "Saved results to cache.");
   }
 
+  verbose && exit(verbose);
 
+  boot;
+}, private=TRUE) # bootstrapSegmentsAndChangepoints()
+
+
+
+setMethodS3("findBootstrapSummaries", "PairedPSCBS", function(fit, what=c("segment", "changepoint"), ..., verbose=FALSE) {
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  # Statistical sanity checks
+  # Validate arguments
   # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-  if (B >= 100L) {
-    verbose && enter(verbose, "Statistical sanity checks (iff B >= 100)");
+  # Argument 'what':
+  what <- match.arg(what);
 
-    # Find extreme quantiles
-    probs <- dimnames(S)[[2]];
-    probs <- gsub("%", "", probs, fixed=TRUE);
-    probs <- as.double(probs) / 100;
+  if (what == "segment") {
+    data <- getSegments(fit);
+  } else if (what == "changepoint") {
+    data <- getChangePoints(fit);
+  }
 
-    # Is it possible to check?
-    if (any(probs < 0.10) && any(probs > 0.90)) {
-      tryCatch({
-        fields <- dimnames(S)[[3]];
-        for (kk in seq(along=fields)) {
-          field <- fields[kk];
-          verbose && enter(verbose, sprintf("Field #%d ('%s') of %d", kk, field, length(fields)));
+  grep("^[^_]+_[^_]+$", colnames(data), value=TRUE);
+}, protected=TRUE) # findBootstrapSummaries()
 
-          # Bootstrap statistics
-          Skk <- S[,,kk, drop=FALSE];
-          dim(Skk) <- dim(Skk)[-3];
 
-          # Sanity checks
-          stopifnot(is.matrix(Skk));
+setMethodS3("hasBootstrapSummaries", "PairedPSCBS", function(fit, ...) {
+  fields <- findBootstrapSummaries(fit, ...);
+  (length(fields) > 0L);
+})
 
-          range <- Skk[,c(1,ncol(Skk)),drop=FALSE];
+setMethodS3("clearBootstrapSummaries", "PairedPSCBS", function(fit, what=c("segment", "changepoint"), ..., verbose=FALSE) {
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Validate arguments
+  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+  # Argument 'what':
+  what <- unique(match.arg(what, several.ok=TRUE));
 
-          # Segmentation means
-          key <- sprintf("%sMean", field);
-          segMean <- segs[[key]];
+  # Argument 'verbose':
+  verbose <- Arguments$getVerbose(verbose);
+  if (verbose) {
+    pushState(verbose);
+    on.exit(popState(verbose));
+  }
 
-          # Segmentation counts
-          cfield <- sprintf("%sNbrOfLoci", ifelse(field == "tcn", "tcn", "dh"));
-          counts <- segs[,cfield,drop=TRUE];
+  verbose && enter(verbose, "Clearing bootstrap summaries");
 
-          if (verbose) {
-            for (rr in seq_len(length(segMean))) {
-              verbose && printf(verbose, "Seg %3d. mean=%g, range=[%g,%g], n=%d\n", rr, segMean[rr], range[rr,1], range[rr,2], counts[rr]);
-            } # for (rr ...)
-          }
+  whats <- what;
+  for (what in whats) {
+    verbose && enter(verbose, sprintf("Clearing %ss", what));
 
-          # Compare only segments with enough data points
-          keep <- (counts > 1L);
-          range <- range[keep,,drop=FALSE];
-          segMean <- segMean[keep];
-
-          # Sanity checks
-          stopifnot(all(range[,2] + tol >= range[,1], na.rm=TRUE));
-          stopifnot(all(segMean + tol >= range[,1], na.rm=TRUE));
-          stopifnot(all(segMean - tol <= range[,2], na.rm=TRUE));
-
-          verbose && exit(verbose);
-        } # for (kk ...)
-      }, error = function(ex) {
-        # If an error, display the data, then throw the exception
-        verbose && cat(verbose, "Tolerance (option 'PSCBS/sanityChecks/tolerance'): ", tol);
-        verbose && print(verbose, segs);
-        throw(ex);
-      })
-    } else {
-      verbose && cat(verbose, "Skipping. Not enough quantiles: ",
-                               paste(dimnames(S)[[2]], collapse=", "));
+    fields <- findBootstrapSummaries(fit, what=what, ...);
+    if (what == "segment") {
+      data <- getSegments(fit);
+      data <- data[,!is.element(colnames(data), fields)];
+      fit$output <- data;
+    } else if (what == "changepoint") {
+      data <- getChangePoints(fit);
+      data <- data[,!is.element(colnames(data), fields)];
+      fit$changepoints <- data;
     }
+    # Sanity check
+    fields <- findBootstrapSummaries(fit, what=what, ...);
+    stopifnot(length(fields) == 0L);
+
+    data <- fields <- NULL; # Not needed anymoew
 
     verbose && exit(verbose);
-  } # if (B >= 100L)
-
-
-  fitB <- fit;
-  fitB$output <- segs;
+  } # for (what ...)
 
   verbose && exit(verbose);
 
-  fitB;
-}, private=TRUE) # bootstrapTCNandDHByRegion()
-
+  fit;
+}, protected=TRUE) # clearBootstrapSummaries()
 
 
 ##############################################################################
 # HISTORY
+# 2013-10-22
+# o SPEEDUP: Added argument 'cache' to bootstrapSegmentsAndChangepoints(),
+#   which caches the results to file if cache=TRUE.
+# 2013-10-21
+# o Added find-, has- and clearBootstrapSummaries().
+# o Added argument 'what' to bootstrapTCNandDHByRegion().
+# o BUG FIX: The new bootstrapSegmentsAndChangepoints() would give
+#   "Error in D[, , 2] : incorrect number of dimensions" if bootstrapping
+#   was done on a single change point (==two segments).
+# 2013-10-20
+# o Now utilizing new getChangePoints().
+# o Now calculating change-point angles in (0,2*pi) using atan2().
+# o Added bootstrapSegmentsAndChangepoints(), which was extract from
+#   internal code of bootstrapTCNandDHByRegion().  The latter now
+#   utilizes the former.
+# o BUG FIX: bootstrapTCNandDHByRegion() did not identify segment
+#   "splitters" as intended.  This has had no impact on the results.
 # 2013-04-23
 # o SPEEDUP: Made bootstrapTCNandDHByRegion() much faster by adding
 #   use.names=FALSE to two internal unlist() statements.
